#include <linux/kernel.h>
#include <linux/module.h>

#include "shfs.h"
#include "shfs_proc.h"
#include "shfs_proto.h"

/* Generic file operations */

char *generic_lsdir = "s_lsdir () { \
if TZ=GMT LC_ALL=POSIX ls -lani \"$s_ROOT$1\" 2>/dev/null; then\
  echo $s_COMPLETE;\
else\
  if test -d \"$s_ROOT$1\"; then\
    echo $s_EPERM;\
  else\
    echo $s_ENOENT;\
  fi \
fi; }\n";

/* Some dd implementations do not like count=0 */

char *generic_read = "s_read () { \
if test \"$3\" = 0; then\
  if test -r \"$s_ROOT$1\"; then\
    echo $s_PRELIM; echo $s_COMPLETE;\
  else\
    echo $s_EPERM; \
  fi \
else\
  if x=`LC_ALL=POSIX dd if=\"$s_ROOT$1\" bs=$4 skip=$5 count=1 2>&1 >/dev/null`; then\
    if echo $s_PRELIM; [ `echo $x|cut -c1-3` = 0+0 ]; then\
      echo $s_COMPLETE;\
    elif dd if=\"$s_ROOT$1\" bs=$4 skip=$5 count=$6 conv=sync 2>/dev/null; then\
      echo $s_COMPLETE;\
    else\
      echo $s_ERROR;\
    fi; \
  else\
    if test -f \"$s_ROOT$1\"; then\
      echo $s_EPERM;\
    else\
      echo $s_ENOENT;\
    fi\
  fi \
fi; }\n";

char *generic_write = "s_write () { \
if test -w \"$s_ROOT$1\"; then\
  echo $s_PRELIM;\
  if test \"$3\" = 0; then\
    echo $s_COMPLETE;\
  else\
   if dd of=\"$s_ROOT$1\" bs=1 seek=$2 count=$3 conv=notrunc 2>/dev/null; then\
      echo $s_COMPLETE;\
    else\
      echo $s_ENOSPC; dd of=/dev/null bs=1 count=$3;\
    fi\
  fi \
else\
  if test -f \"$s_ROOT$1\"; then\
    echo $s_EPERM; \
  else\
    echo $s_ENOENT;\
  fi; \
fi; }\n";

char *generic_mkdir = "s_mkdir () { \
if mkdir \"$s_ROOT$1\" 2>/dev/null; then\
  echo $s_COMPLETE; \
else\
  echo $s_EPERM; \
fi; }\n";

char *generic_rmdir = "s_rmdir () { \
if test -d \"$s_ROOT$1\"; then\
  if rmdir \"$s_ROOT$1\" 2>/dev/null; then\
    echo $s_COMPLETE;\
  else\
    echo $s_EPERM;\
  fi \
else\
  echo $_ENOENT; \
fi; }\n";

char *generic_mv = "s_mv () { \
if mv -f \"$s_ROOT$1\" \"$s_ROOT$2\" 2>/dev/null; then\
  echo $s_COMPLETE; \
else\
  if test -f \"$s_ROOT$1\"; then\
    echo $s_EPERM;\
  else\
    echo $s_ENOENT;\
  fi \
fi; }\n";

char *generic_rm = "s_rm () { \
if rm -f \"$s_ROOT$1\" 2>/dev/null; then\
  echo $s_COMPLETE; \
else\
  if test -f \"$s_ROOT$1\"; then\
    echo $s_EPERM;\
  else\
    echo $s_ENOENT;\
  fi \
fi; }\n";

char *generic_creat = "s_creat () { \
if ( >\"$s_ROOT$1\"; chmod $2 \"$s_ROOT$1\" ) 2>/dev/null; then\
  echo $s_COMPLETE; \
else\
  echo $s_EPERM; \
fi; }\n";

char *generic_ln = "s_ln () { \
if ln -f \"$s_ROOT$1\" \"$s_ROOT$2\" 2>/dev/null; then\
  echo $s_COMPLETE; \
else\
  echo $s_EPERM; \
fi; }\n";

char *generic_sln = "s_sln () { \
if ln -s -f \"$1\" \"$s_ROOT$2\" 2>/dev/null; then\
  echo $s_COMPLETE; \
else\
  echo $s_EPERM; \
fi; }\n";

char *generic_chmod = "s_chmod () { \
if chmod $2 \"$s_ROOT$1\" 2>/dev/null; then\
  echo $s_COMPLETE; \
else\
  if test -f \"$s_ROOT$1\"; then\
    echo $s_EPERM;\
  else\
    echo $s_ENOENT;\
  fi \
fi; }\n";

char *generic_chown = "s_chown () { \
if chown $2 \"$s_ROOT$1\" 2>/dev/null; then\
  echo $s_COMPLETE; \
else\
  if test -f \"$s_ROOT$1\"; then\
    echo $s_EPERM;\
  else\
    echo $s_ENOENT;\
  fi \
fi; }\n";

char *generic_chgrp = "s_chgrp () { \
if chgrp $2 \"$s_ROOT$1\" 2>/dev/null; then\
  echo $s_COMPLETE; \
else\
  if test -f \"$s_ROOT$1\"; then\
    echo $s_EPERM;\
  else\
    echo $s_ENOENT;\
  fi \
fi; }\n";

/* some dd implementations do not like count=0 */

char *generic_trunc = "s_trunc () { \
if test \"$3\" = 0; then\
  if ( >\"$s_ROOT$1\" ) 2>/dev/null; then\
    echo $s_COMPLETE; \
  else\
    if test -f \"$s_ROOT$1\"; then\
      echo $s_EPERM;\
    else\
      echo $s_ENOENT;\
    fi\
  fi \
else\
  if dd if=/dev/zero of=\"$s_ROOT$1\" bs=$2 seek=$3 count=0 2>/dev/null; then\
    echo $s_COMPLETE; \
  else\
    if test -f \"$s_ROOT$1\"; then\
      echo $s_EPERM;\
    else\
      echo $s_ENOENT;\
    fi\
  fi \
fi; }\n";

/* Linux specific operations */

char *linux_read = "s_read () { \
if x=`LC_ALL=POSIX dd if=\"$s_ROOT$1\" bs=1 skip=$2 count=1 2>&1 >/dev/null`; then\
  if echo $s_PRELIM; test `echo $x|cut -c1-3` = 0+0; then\
    echo $s_COMPLETE;\
  elif dd if=\"$s_ROOT$1\" bs=$4 skip=$5 count=$6 conv=sync 2>/dev/null; then\
    echo $s_COMPLETE;\
  else\
    echo $s_ERROR;\
  fi; \
else\
  if test -e \"$s_ROOT$1\"; then\
    echo $s_EPERM;\
  else\
    echo $s_ENOENT;\
  fi \
fi; }\n";

/* We should have head command (OSF1 has one too) */

char *linux_write = "s_write () { \
if printf '' >\"$s_ROOT$1\"._shfs_ 2>/dev/null; then\
  echo $s_PRELIM;\
  if head -c $3 >\"$s_ROOT$1\"._shfs_ 2>/dev/null; then\
    if dd if=\"$s_ROOT$1\"._shfs_ of=\"$s_ROOT$1\" bs=$4 seek=$5 conv=notrunc 2>/dev/null; then\
      echo $s_COMPLETE;\
    else\
      if test -e \"$s_ROOT$1\"; then\
        echo $s_EPERM;\
      else\
        echo $s_ENOENT;\
      fi\
    fi;\
  else\
    echo $s_ENOSPC; dd of=/dev/null bs=1 count=$3;\
  fi;\
  rm -f \"$s_ROOT$1\"._shfs_ 2>/dev/null; \
else\
  echo $s_EPERM; \
fi; }\n";

/* more specific first! */

struct shfs_system_t shfs_system[] = {
	{
		"Linux", NULL, {
			&generic_lsdir,
			&linux_read,
			&linux_write,
			&generic_mkdir,
			&generic_rmdir,
			&generic_mv,
			&generic_rm,
			&generic_creat,
			&generic_ln,
			&generic_sln,
			&generic_chmod,
			&generic_chown,
			&generic_chgrp,
			&generic_trunc,
		}
	}, {
		"OSF1", NULL, {
			&generic_lsdir,
			&generic_read,
			&linux_write,
			&generic_mkdir,
			&generic_rmdir,
			&generic_mv,
			&generic_rm,
			&generic_creat,
			&generic_ln,
			&generic_sln,
			&generic_chmod,
			&generic_chown,
			&generic_chgrp,
			&generic_trunc,
		}
	}, {
		NULL, NULL, {
			&generic_lsdir,
			&generic_read,
			&generic_write,
			&generic_mkdir,
			&generic_rmdir,
			&generic_mv,
			&generic_rm,
			&generic_creat,
			&generic_ln,
			&generic_sln,
			&generic_chmod,
			&generic_chown,
			&generic_chgrp,
			&generic_trunc,
		}
	}
};

/*

Protocol NOTES:

* All operations should redirect stderr to /dev/null.
* You *MUST* quote strings with single quotes (expansion ' -> '\'' 
  is done in filenames).
* Read & write must work correctly when count == 0 (return zero bytes)
 
*/

/* Machine independent operations */

int 
do_uname(struct shfs_sb_info *info)
{
	return shfs_pipe_printf(info, "if LC_ALL=POSIX uname -s -r 2>/dev/null; then echo '### %d'; else echo '### %d'; fi\n", REP_COMPLETE, REP_ERROR);
}

int 
do_ping(struct shfs_sb_info *info, unsigned long seq)
{
	return shfs_pipe_printf(info, "\n\necho; echo '### %d'; echo %lu; echo '### %d'\n", REP_PRELIM, seq, REP_NOP);
}

int
do_init(struct shfs_sb_info *info, int sys)
{
	int res, r;
	char **s;
	int i;
	
	res = shfs_pipe_printf(info,\
"s_ROOT='%s';\n"
"s_PRELIM='### %d'; s_COMPLETE='### %d'; s_NOP='### %d';\n"
"s_CONTINUE='### %d'; s_TRANSIENT='### %d';\n"
"s_ERROR='### %d'; s_EPERM='### %d'; s_ENOSPC='### %d'; s_ENOENT='### %d';\n",
info->mnt.root,\
REP_PRELIM, REP_COMPLETE, REP_NOP,\
REP_CONTINUE, REP_TRANSIENT,\
REP_ERROR, REP_EPERM, REP_ENOSPC, REP_ENOENT);

	if (res < 0)
		return res;
	
	for (i = 0; i < SHFS_OPS; i++) {
		s = shfs_system[sys].ops[i];
		r = shfs_pipe_printf(info, *s);
		if (r < 0)
			return r;
		res += r;
	}

	r = shfs_pipe_printf(info, "echo '### %d'\n", REP_COMPLETE);
	if (r < 0)
		return r;
	
	return res+r;
}

